"""Custom sklearn transformers
"""
import logging
from typing import List, Tuple, Any, Union
from string import punctuation
from nltk.corpus import stopwords

import pandas as pd
import numpy as np

from sklearn.base import TransformerMixin
from sklearn.feature_extraction.text import CountVectorizer

import nltk
from nltk.stem import WordNetLemmatizer
from nltk.stem.lancaster import LancasterStemmer
from nltk import word_tokenize, pos_tag
from nltk.sentiment.vader import SentimentIntensityAnalyzer


class Debug(TransformerMixin):
    """Debug transformer
    """

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:  # pylint: disable=no-self-use
        """Debug transformer
        """
        print(df.shape)
        print(df)
        return df


class SentimentAnalyzer(TransformerMixin):
    """ Create sentiment analyzer on text data

    """

    def __init__(self, col: str):
        nltk.download('vader_lexicon')
        self.col = col
        self.analyzer = SentimentIntensityAnalyzer()
        self.wrong_input = False

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    def _analyze_sentiment(self, text: str) -> pd.DataFrame:
        #log only the first input
        sentiments = self.analyzer.polarity_scores(str(text))
        return sentiments

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as sentiment score

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with one new column by columns in self.cols.
        """
        sentiment = df.loc[:, self.col].map(self._analyze_sentiment)
        sentiment = pd.DataFrame(list(sentiment))

        # sentiment = sentiment.drop(self.col, axis=1)
        sentiment.columns = [f"{x}_{self.col}" for x in sentiment.columns]

        return pd.concat([df, sentiment.set_index(df.index)], axis=1)


class FlairHandler(TransformerMixin):
    """ Create binary variable from flair

    """

    def __init__(self, col: str):
        logging.warning("Using FlairHandler: could lead to overfitting")
        self.col = col

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as length of fielsds

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with one new column by columns in self.cols.
        """
        nopizza = df.loc[:, self.col].map(lambda x: x is None).astype(int)
        receivedonly = df.loc[:, self.col].map(lambda x: x == "shroom").astype(int)
        receivedandgiven = df.loc[:, self.col].map(lambda x: x == "PIF").astype(int)
        new_columns = pd.concat([nopizza, receivedonly, receivedandgiven], axis=1)
        new_columns.columns = ["nopizzareceived", "receivedonly", "receivedandgiven"]
        return pd.concat([df, new_columns], axis=1)


class DateToInt(TransformerMixin):
    """ Create a new date column as int type
    Attributes
    ----------
    cols: list of columns

    """

    def __init__(self, cols: List[str]):
        self.cols = cols

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as length of fielsds

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with one new column by columns in self.cols.
        """
        for col in self.cols:
            df.loc[:, f"{col}_INT"] = df.loc[:, col].astype(int)
        return df


class LengthOfField(TransformerMixin):
    """ Create new columns as length of field for text or list
    Attributes
    ----------
    cols: list of columns

    """

    def __init__(self, cols: List[str]):
        self.cols = cols

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    @staticmethod
    def _len(x: Union[str, List[Any]], col: str):
        try:
            return len(x)
        except TypeError as mytypeerror:
            raise TypeError(
                f"Found object {x} of type {type(x)} in column {col}, expected str or list type") from mytypeerror

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as length of fielsds

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with one new column by columns in self.cols.
        """
        for col in self.cols:
            df.loc[:, f"{col}_LENGTH"] = df.loc[:, col].map(lambda x: self._len(x, col))  #pylint: disable=cell-var-from-loop
        return df


class DiffWithMax(TransformerMixin):
    """ Create new column as difference between the max found and the value
    Attributes
    ----------
    cols: list of columns

    """

    def __init__(self, cols: List[str], date_as_int: bool = True):
        self.cols = cols
        self.max = pd.Series()
        self.date_as_int = date_as_int

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Compute the max and keep it
        """
        self.max = df.loc[:, self.cols].max().astype(float)
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as diff with max

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with new column as division of others columns.
        """
        for col in self.cols:
            df.loc[:, f"{col}_DIFF_WITH_MAX"] = self.max[col] - df.loc[:, col].astype(float)
        return df


class DiffBetweenCols(TransformerMixin):
    """ Create new column as difference between two columns
    Attributes
    ----------
    cols_tuple: list of string tuple (a, b) such as result will be a / b

    """

    def __init__(self, cols_tuple: List[Tuple[str]]):
        self.cols_tuple = cols_tuple

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as division of others

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with new column as division of others columns.
        """
        for col_a, col_b in self.cols_tuple:  # type: ignore
            df.loc[:, str(col_a) + "_SUBSTRACTED_BY_" +  # type: ignore
                   str(col_b)] = df.loc[:, col_a].astype(float) - df.loc[:, col_b].astype(float)  # type: ignore
        return df


class DivBetweenCols(TransformerMixin):
    """ Create new column as division between two columns
    Attributes
    ----------
    cols_tuple: list of string tuple (a, b) such as result will be a / b

    """

    def __init__(self, cols_tuple: List[Tuple[str]]):
        self.cols_tuple = cols_tuple

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:
        """Create new column as division of others

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with new column as division of others columns.
        """
        for col_a, col_b in self.cols_tuple:  # type: ignore
            df.loc[:, str(col_a) + "_DIVIDED_BY_" + str(col_b)] = pd.to_numeric(  # type: ignore
                df.loc[:, col_a], errors='coerce') / pd.to_numeric(df.loc[:, col_b], errors='coerce')  # type: ignore
        return df


class KeywordFoundExtraction(TransformerMixin):
    """ extract keyword boolean from text.
    """

    def __init__(self, col: str, keywords: List[str]):
        self.col = col
        self.keywords = keywords

    def fit(self, df, y=None, **fit_params):
        """ Empty fit
        """
        return self

    @staticmethod
    def _keyword_in_str(keyword: str, text: str) -> bool:
        return keyword in text

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """ Add bolean to pandas dataframe if keywoard are found

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with new columns.
        """
        logging.warning("KeywordExtraction is probably overfitting")
        text_col = df.loc[:, self.col]
        new_columns = []
        for key in self.keywords:
            new_columns.append(text_col.map(lambda x: self._keyword_in_str(key, x)))  #pylint: disable=cell-var-from-loop
        new_df_columns = pd.concat(new_columns, axis=1)
        new_df_columns.columns = [f"{self.col}_{key}_found" for key in self.keywords]
        return pd.concat([df, new_df_columns], axis=1)


class PreprocessString(TransformerMixin):
    """ Preprocess string columns
    """

    def __init__(self, cols: List[str]):
        nltk.download('stopwords')
        nltk.download('punkt')
        nltk.download('averaged_perceptron_tagger')
        nltk.download('wordnet')
        nltk.download('omw-1.4')

        self.cols = cols
        self.wnl = LancasterStemmer()

    def fit(self, df, y=None, **fit_params):
        """ Empty fit
        """
        return self

    def _preprocess(self, input_string: str) -> str:
        res = " ".join([self.wnl.stem(word.lower()) for word, tag in pos_tag(word_tokenize(input_string))])
        return res

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Preprocess text columns

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with text columns preprocessed.
        """
        df.loc[:, self.cols] = df.loc[:, self.cols].applymap(self._preprocess)
        return df


class ProcessText(TransformerMixin):
    """ Class used to process text
    """

    def __init__(self, text_column: str):
        # Loading NLTK if we use the transformers
        nltk.download('stopwords')
        nltk.download('punkt')
        nltk.download('averaged_perceptron_tagger')
        nltk.download('wordnet')
        nltk.download('omw-1.4')

        self.text_column = text_column

        self.stopwords = self._get_stop_list()
        self.wnl = WordNetLemmatizer()

        self.count_vectorizer = CountVectorizer(analyzer=self._preprocess_text)

    @staticmethod
    def _get_stop_list():
        stopwords_json = {
            "en": [
                "a", "a's", "able", "about", "above", "according", "accordingly", "across", "actually", "after",
                "afterwards", "again", "against", "ain't", "all", "allow", "allows", "almost", "alone", "along",
                "already", "also", "although", "always", "am", "among", "amongst", "an", "and", "another", "any",
                "anybody", "anyhow", "anyone", "anything", "anyway", "anyways", "anywhere", "apart", "appear",
                "appreciate", "appropriate", "are", "aren't", "around", "as", "aside", "ask", "asking", "associated",
                "at", "available", "away", "awfully", "b", "be", "became", "because", "become", "becomes", "becoming",
                "been", "before", "beforehand", "behind", "being", "believe", "below", "beside", "besides", "best",
                "better", "between", "beyond", "both", "brief", "but", "by", "c", "c'mon", "c's", "came", "can",
                "can't", "cannot", "cant", "cause", "causes", "certain", "certainly", "changes", "clearly", "co", "com",
                "come", "comes", "concerning", "consequently", "consider", "considering", "contain", "containing",
                "contains", "corresponding", "could", "couldn't", "course", "currently", "d", "definitely", "described",
                "despite", "did", "didn't", "different", "do", "does", "doesn't", "doing", "don't", "done", "down",
                "downwards", "during", "e", "each", "edu", "eg", "eight", "either", "else", "elsewhere", "enough",
                "entirely", "especially", "et", "etc", "even", "ever", "every", "everybody", "everyone", "everything",
                "everywhere", "ex", "exactly", "example", "except", "f", "far", "few", "fifth", "first", "five",
                "followed", "following", "follows", "for", "former", "formerly", "forth", "four", "from", "further",
                "furthermore", "g", "get", "gets", "getting", "given", "gives", "go", "goes", "going", "gone", "got",
                "gotten", "greetings", "h", "had", "hadn't", "happens", "hardly", "has", "hasn't", "have", "haven't",
                "having", "he", "he's", "hello", "help", "hence", "her", "here", "here's", "hereafter", "hereby",
                "herein", "hereupon", "hers", "herself", "hi", "him", "himself", "his", "hither", "hopefully", "how",
                "howbeit", "however", "i", "i'd", "i'll", "i'm", "i've", "ie", "if", "ignored", "immediate", "in",
                "inasmuch", "inc", "indeed", "indicate", "indicated", "indicates", "inner", "insofar", "instead",
                "into", "inward", "is", "isn't", "it", "it'd", "it'll", "it's", "its", "itself", "j", "just", "k",
                "keep", "keeps", "kept", "know", "known", "knows", "l", "last", "lately", "later", "latter", "latterly",
                "least", "less", "lest", "let", "let's", "like", "liked", "likely", "little", "look", "looking",
                "looks", "ltd", "m", "mainly", "many", "may", "maybe", "me", "mean", "meanwhile", "merely", "might",
                "more", "moreover", "most", "mostly", "much", "must", "my", "myself", "n", "name", "namely", "nd",
                "near", "nearly", "necessary", "need", "needs", "neither", "never", "nevertheless", "new", "next",
                "nine", "no", "nobody", "non", "none", "noone", "nor", "normally", "not", "nothing", "novel", "now",
                "nowhere", "o", "obviously", "of", "off", "often", "oh", "ok", "okay", "old", "on", "once", "one",
                "ones", "only", "onto", "or", "other", "others", "otherwise", "ought", "our", "ours", "ourselves",
                "out", "outside", "over", "overall", "own", "p", "particular", "particularly", "per", "perhaps",
                "placed", "please", "plus", "possible", "presumably", "probably", "provides", "q", "que", "quite", "qv",
                "r", "rather", "rd", "re", "really", "reasonably", "regarding", "regardless", "regards", "relatively",
                "respectively", "right", "s", "said", "same", "saw", "say", "saying", "says", "second", "secondly",
                "see", "seeing", "seem", "seemed", "seeming", "seems", "seen", "self", "selves", "sensible", "sent",
                "serious", "seriously", "seven", "several", "shall", "she", "should", "shouldn't", "since", "six", "so",
                "some", "somebody", "somehow", "someone", "something", "sometime", "sometimes", "somewhat", "somewhere",
                "soon", "sorry", "specified", "specify", "specifying", "still", "sub", "such", "sup", "sure", "t",
                "t's", "take", "taken", "tell", "tends", "th", "than", "thank", "thanks", "thanx", "that", "that's",
                "thats", "the", "their", "theirs", "them", "themselves", "then", "thence", "there", "there's",
                "thereafter", "thereby", "therefore", "therein", "theres", "thereupon", "these", "they", "they'd",
                "they'll", "they're", "they've", "think", "third", "this", "thorough", "thoroughly", "those", "though",
                "three", "through", "throughout", "thru", "thus", "to", "together", "too", "took", "toward", "towards",
                "tried", "tries", "truly", "try", "trying", "twice", "two", "u", "un", "under", "unfortunately",
                "unless", "unlikely", "until", "unto", "up", "upon", "us", "use", "used", "useful", "uses", "using",
                "usually", "uucp", "v", "value", "various", "very", "via", "viz", "vs", "w", "want", "wants", "was",
                "wasn't", "way", "we", "we'd", "we'll", "we're", "we've", "welcome", "well", "went", "were", "weren't",
                "what", "what's", "whatever", "when", "whence", "whenever", "where", "where's", "whereafter", "whereas",
                "whereby", "wherein", "whereupon", "wherever", "whether", "which", "while", "whither", "who", "who's",
                "whoever", "whole", "whom", "whose", "why", "will", "willing", "wish", "with", "within", "without",
                "won't", "wonder", "would", "wouldn't", "x", "y", "yes", "yet", "you", "you'd", "you'll", "you're",
                "you've", "your", "yours", "yourself", "yourselves", "z", "zero"
            ]
        }
        stopwords_json_en = set(stopwords_json['en'])
        stopwords_nltk_en = set(stopwords.words('english'))
        stopwords_punct = set(punctuation)
        # Combine the stopwords. Its a lot longer so I'm not printing it out...
        stoplist_combined = set.union(stopwords_json_en, stopwords_nltk_en, stopwords_punct)
        return stoplist_combined

    def _lemmatize_sent(self, text):

        def penn2morphy(penntag):
            """ Converts Penn Treebank tags to WordNet. """
            morphy_tag = {'NN': 'n', 'JJ': 'a', 'VB': 'v', 'RB': 'r'}
            try:
                return morphy_tag[penntag[:2]]
            # Wrong: using except without Exception type
            except:  #pylint: disable = bare-except
                return 'n'

        # Text input is string, returns lowercased strings.
        return [self.wnl.lemmatize(word.lower(), pos=penn2morphy(tag)) for word, tag in pos_tag(word_tokenize(text))]

    def _preprocess_text(self, text: str):
        # Input: str, i.e. document/sentence
        # Output: list(str) , i.e. list of lemmas
        return [word for word in self._lemmatize_sent(text) if word not in self.stopwords and not word.isdigit()]

    def fit(self, df, y=None, **fit_params):
        """fit
        """
        # pylint: disable=unused-argument
        self.count_vectorizer.fit(df.loc[:, self.text_column])
        return self

    def transform(self, df, **transform_params):
        """ Transform
        """
        # pylint: disable=unused-argument
        return self.count_vectorizer.transform(df.loc[:, self.text_column])


class SklearnPandasWrapper(TransformerMixin):
    """ Class used to wrap SKLEARN transformer to use them
    but keep pandas format and columns name
    """

    def __init__(self, transformer):
        # self.__class__ = type(transformer.__class__.__name__,
        #                       (self.__class__, transformer.__class__), {})
        self.__dict__ = transformer.__dict__
        self.transformer = transformer

    def fit(self, df, y=None, **fit_params):
        """fit
        """
        # pylint: disable=unused-argument
        self.transformer.fit(df, fit_params)
        return self

    def transform(self, df, **transform_params):
        """ Transform
        """
        # pylint: disable=unused-argument
        if transform_params:
            res = self.transformer.transform(df, transform_params)
        else:
            res = self.transformer.transform(df)
        res = pd.DataFrame(res, columns=df.columns)
        return res


class GetFeaturesNames(TransformerMixin):
    """Transformer used only to get feature name in a pipeline step
    """

    def __init__(self):
        self.features_names = None

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """Set feature name to attributes
        """
        self.feature_names = df.columns  #pylint: disable=attribute-defined-outside-init
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:  # pylint: disable=no-self-use
        """Empty transform
        """
        return df


class SelectColumns(TransformerMixin):
    """Select a subset of columns in X
    """

    def __init__(self, cols: List[str] = None):
        self.cols = cols if cols else []

    def fit(self, df, y=None, **fit_params):
        """ Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Select columns.

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with only selected columns.
        """
        return df.loc[:, self.cols]


class DropColumns(TransformerMixin):
    """Drop columns in X
    """

    def __init__(self, cols: List[str] = None):
        self.cols = cols if cols else []

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """ Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Drop columns.

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with the columns droped.
        """
        return df.drop(self.cols, axis=1, errors="ignore")


class UnixTimeStampToPandasDate(TransformerMixin):
    """Convert columns with unix time stamp to pandas datetime
    """

    def __init__(self, cols: List[str] = None):
        self.cols = cols if cols else []

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """ Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame) -> pd.DataFrame:
        """Drop columns.

        Args:
            df (pd.DataFrame): input dataframe

        Returns:
            pd.DataFrame: Output dataframe with the columns droped.
        """
        for col in self.cols:
            df.loc[:, col] = pd.to_datetime(df.loc[:, col])
        return df


class DateFeatureGeneration(TransformerMixin):
    """Feature Generation for time series

    Parameters
    ----
    col_date: (str) name of column with date

    Attributes
    ----
    Return pandas dataframe with new columns generated
    """

    def __init__(self, col_date: str, dummify: bool = True):
        self.col_date = col_date
        self.dummify = dummify

    def fit(self, df: pd.DataFrame, y: pd.DataFrame = None, **fit_params):
        """ Empty fit
        """
        return self

    def transform(self, df: pd.DataFrame, **transform_params) -> pd.DataFrame:  # pylint: disable=too-many-locals
        """Create multiple variable from a date column.

        Args:
            df (pd.DataFrame): Pandas dataframe with date column

        Returns:
            pd.DataFrame: input df with new columns.
        """
        date_series = df.loc[:, self.col_date]
        month = date_series.map(lambda x: x.month)

        weekday = date_series.map(lambda x: x.weekday())
        hour = date_series.map(lambda x: x.hour)
        year = date_series.map(lambda x: x.year)
        day = date_series.map(lambda x: x.day)

        # Add cyclical feature transformation
        hr_sin = np.sin(hour * (2. * np.pi / 24))
        hr_cos = np.cos(hour * (2. * np.pi / 24))
        mnth_sin = np.sin(month * (2. * np.pi / 12))
        mnth_cos = np.cos(month * (2. * np.pi / 12))
        weekday_sin = np.sin(weekday * (2. * np.pi / 7))
        weekday_cos = np.cos(weekday * (2. * np.pi / 7))

        x_date = pd.concat(
            [month, weekday, hour, year, day, hr_sin, hr_cos, mnth_sin, mnth_cos, weekday_sin, weekday_cos], axis=1)

        dummy_columns = [self.col_date + "_" + x for x in ["month", "weekday", "hour", "year", "day"]]
        others_columns = [
            self.col_date + "_" + x for x in ["hr_sin", "hr_cos", "mnth_sin", "mnth_cos", "weekday_sin", "weekday_cos"]
        ]

        new_colnames = dummy_columns + others_columns

        x_date.columns = new_colnames
        if self.dummify is True:
            dummy_date = pd.get_dummies(x_date, dummy_na=True, columns=dummy_columns)
            res = pd.concat([df, dummy_date], axis=1)
        else:
            for var in x_date.columns:
                x_date.loc[:, var] = x_date.loc[:, var].astype("category")
            res = pd.concat([df, x_date], axis=1)
        # Others variable
        res.loc[:, self.col_date + "_" + "is_end_month"] =\
            date_series.dt.is_month_end
        res.loc[:, self.col_date + "_" + "is_start_month"] =\
            date_series.dt.is_month_start
        res.loc[:, self.col_date + "_" + "is_start_quarter"] =\
            date_series.dt.is_quarter_start
        res.loc[:, self.col_date + "_" + "is_end_quarter"] =\
            date_series.dt.is_quarter_end
        res.loc[:, self.col_date + "_" + "is_end_year"] =\
            date_series.dt.is_year_end
        res.loc[:, self.col_date + "_" + "is_start_year"] =\
            date_series.dt.is_year_start
        res.loc[:, self.col_date + "_" + "is_leap_year"] =\
            date_series.dt.is_leap_year
        return res
